// BigInteger

package java.math;

import java.util.Random;

//public class BigInteger extends Number implements Comparable {
public class BigInteger {

	private int sign;	// -1 means -ve; +1 means +ve; 0 means 0;
	private int magnitude[];// array of ints with [0] being the most significant
	private int nBits = -1; // cache bitCount() value
	private int nBitLength = -1; // cache bitLength() value
	private static final long IMASK = 0xffffffffL;

    private BigInteger()
    {
    }

	private BigInteger(int nWords) {
		sign = 1;
		magnitude = new int[nWords];
	}


	private BigInteger(
		int signum,
		int[] mag)
	{
		sign = signum;
		if (mag.length > 0) {
			int i = 0;
			while (i < mag.length && mag[i] == 0) i++;
			if (i == 0) {
				magnitude = mag;
				}
			else {
				// strip leading 0 bytes
				int[] newMag = new int[mag.length - i];
				System.arraycopy(mag, i, newMag, 0, newMag.length);
				magnitude = newMag;
				if (newMag.length == 0) sign = 0;
				}
			}
		else {
			magnitude = mag;
			sign = 0;
			}
	}


	public BigInteger(
	    String sval)
		throws NumberFormatException
	{
		this(sval, 10);
	}


	public BigInteger(
	    String sval,
		int rdx)
		throws NumberFormatException
	{
		if (sval.length() == 0) {
			throw new NumberFormatException("Zero length BigInteger");
			}

		if (rdx < Character.MIN_RADIX || rdx > Character.MAX_RADIX) {
			throw new NumberFormatException("Radix out of range");
			}

		int index = 0;
		sign = 1;

		if (sval.charAt(0) == '-') {
			if (sval.length() == 1) {
				throw new NumberFormatException("Zero length BigInteger");
				}

			sign = -1;
			index = 1;
			}

		// strip leading zeros from the string value
		while (index < sval.length() &&
				Character.digit(sval.charAt(index), rdx) == 0) {
			index++;
			}

		if (index >= sval.length()) {
			// zero value - we're done
			sign = 0;
			magnitude = new int[0];
			return;
			}

		//////
		// could we work out the max number of ints required to store
		// sval.length digits in the given base, then allocate that
		// storage in one hit?, then generate the magnitude in one hit too?
		//////

		BigInteger b = BigInteger.ZERO;
		BigInteger r = valueOf(rdx);
		while (index < sval.length()) {
			// (optimise this by taking chunks of digits instead?)
			b = b.multiply(r).add(valueOf(Character.digit(sval.charAt(index), rdx)));
			index++;
			}

		magnitude = b.magnitude;

		return;
	}


	public BigInteger(
	    byte[] bval)
		throws NumberFormatException
	{
		if (bval.length == 0) {
			throw new NumberFormatException("Zero length BigInteger");
			}

		sign = 1;
		if (bval[0] < 0) {
			// FIXME:
			int iBval;
			sign = -1;
			// strip leading sign bytes
			for (iBval = 0; iBval < bval.length && bval[iBval] == -1; iBval++) ;
			magnitude = new int[(bval.length - iBval)/2 + 1];
			// copy bytes to magnitude
			// invert bytes then add one to find magnitude of value
			}
		else {
			// strip leading zero bytes and return magnitude bytes
			magnitude = makeMagnitude(bval);
			}
	}


	private int[] makeMagnitude(byte[] bval) {
		int i;
		int[] mag;
		int firstSignificant;

		// strip leading zeros
		for (firstSignificant = 0;
			firstSignificant < bval.length && bval[firstSignificant] == 0;
			firstSignificant++) ;

		if (firstSignificant >= bval.length) {
			return new int[0];
			}

		int nInts = (bval.length - firstSignificant + 3) / 4;
		int bCount = (bval.length - firstSignificant) % 4;
		if (bCount == 0) bCount = 4;

		mag = new int[nInts];
		int v = 0;
		int magnitudeIndex = 0;
		for (i = firstSignificant; i < bval.length; i++) {
			v <<= 8;
			v |= ((int)(bval[i])) & 0xff;
			bCount--;
			if (bCount <= 0) {
				mag[magnitudeIndex] = v;
				magnitudeIndex++;
				bCount = 4;
				v = 0;
				}
			}

		if (magnitudeIndex < mag.length) {
			mag[magnitudeIndex] = v;
			}

		return mag;
	}


	public BigInteger(
		int sign,
		byte[] mag)
		throws NumberFormatException
	{
		if (sign < -1 || sign > 1) {
			throw new NumberFormatException("Invalid sign value");
			}

		if (sign == 0) {
			this.sign = 0;
			this.magnitude = new int[0];
			return;
			}

		// copy bytes
		this.magnitude = makeMagnitude(mag);
		this.sign = sign;
	}


	public BigInteger(
	    int numBits,
		Random rnd)
		throws IllegalArgumentException
	{
		if (numBits < 0) {
			throw new IllegalArgumentException("numBits must be non-negative");
			}

		int nBytes = (numBits + 7) / 8;

		byte[] b = new byte[nBytes];

        do
        {
		    if (nBytes > 0)
            {
			    nextRndBytes(rnd, b);
			    // strip off any excess bits in the MSB
			    b[0] &= rndMask[8*nBytes - numBits];
			}

            this.magnitude = makeMagnitude(b);
            this.sign = 1;
	        this.nBits = -1;
	        this.nBitLength = -1;
        }
        while (this.bitLength() != numBits);
	}


	private static final int BITS_PER_BYTE = 8;
	private static final int BYTES_PER_INT = 4;

    /**
     * strictly speaking this is a little dodgey from a compliance
     * point of view as it forces people to be using SecureRandom as
     * well, that being said - this implementation is for a crypto
     * library and you do have the source!
     */
	private void nextRndBytes(
        Random rnd,
        byte[] bytes)
    {
		int numRequested = bytes.length;
		int numGot = 0, r = 0;

        if (rnd instanceof java.security.SecureRandom)
        {
            ((java.security.SecureRandom)rnd).nextBytes(bytes);
        }
        else
        {
            for (;;)
            {
			    for (int i = 0; i < BYTES_PER_INT; i++)
                {
                    if (numGot == numRequested)
                    {
                        return;
                    }

                    r = (i==0 ? rnd.nextInt() : r >> BITS_PER_BYTE);
				    bytes[numGot++] = (byte)r;
				}
			}
        }
	}

	private static final byte[] rndMask = {(byte)255, 127, 63, 31, 15, 7, 3, 1};


	public BigInteger(
	    int bitLength,
		int certainty,
		Random rnd)
		throws ArithmeticException
	{
		BigInteger biggie;

		do
        {
            biggie = new BigInteger(bitLength, rnd);
            if (certainty > 0 && bitLength > 2)
            {
                biggie.magnitude[biggie.magnitude.length - 1] |= 1;
            }
        }
        while (!biggie.isProbablePrime(certainty));

		sign = 1;
		this.magnitude = biggie.magnitude;
	}


	private int getMagInt(
		int index)
	{
		if (index >= this.magnitude.length || index < 0) {
			return (this.sign >= 0) ? 0 : -1;
			}

		return magnitude[index];
	}


	public BigInteger abs()
	{
		return (sign >= 0) ? this : this.negate();
	}


	public BigInteger add(
		BigInteger val)
		throws ArithmeticException
	{
		if (val.sign == 0 || val.magnitude.length == 0) return this;
		if (this.sign == 0 || this.magnitude.length == 0) return val;

		if (val.sign < 0)
        {
			if (this.sign > 0) return this.subtract(val.negate());
        }
		else
        {
			if (this.sign < 0) return val.subtract(this.negate());
        }

		// both BigIntegers are either +ve or -ve; set the sign later

		//**** size of new int vector should be max of either source
		//**** vector + 1 more if overflow can occur via MSB add
		int mag[] = new int[max(val.magnitude.length, this.magnitude.length) + 1];
		int rI = mag.length - 1;
		int vI = val.magnitude.length - 1;
		int tI = this.magnitude.length - 1;
		long m = 0;

        if (vI < tI)
        {
            while (vI >= 0)
            {
                m += (((long)this.magnitude[tI--]) & IMASK)
                    + (((long)val.magnitude[vI--]) & IMASK);
                mag[rI--] = (int)m;
                m >>>= 32;
            }

            while (tI >= 0)
            {
                m += (((long)this.magnitude[tI--]) & IMASK);
                mag[rI--] = (int)m;
                m >>>= 32;
            }

            mag[rI] = (int)m;
        }
        else
        {
            while (tI >= 0)
            {
                m += (((long)this.magnitude[tI--]) & IMASK)
                    + (((long)val.magnitude[vI--]) & IMASK);
                mag[rI--] = (int)m;
                m >>>= 32;
            }

            while (vI >= 0)
            {
                m += (((long)val.magnitude[vI--]) & IMASK);
                mag[rI--] = (int)m;
                m >>>= 32;
            }

            mag[rI] = (int)m;
        }

		return new BigInteger(this.sign, mag);
	}


	public int bitCount()
	{
		if (nBits == -1) {
			nBits = 0;
			for (int i = 0; i < magnitude.length; i++) {
				nBits += bitCounts[magnitude[i] & 0xff];
				nBits += bitCounts[(magnitude[i] >> 8) & 0xff];
				nBits += bitCounts[(magnitude[i] >> 16) & 0xff];
				nBits += bitCounts[(magnitude[i] >> 24) & 0xff];
				}
			}

		return nBits;
	}


    private final static byte bitCounts[] = {
		0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4,
		1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
		1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
		2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
		1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
		2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
		2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
		3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
		1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
		2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
		2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
		3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
		2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
		3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
		3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
		4, 5, 5, 6, 5, 6, 6, 7, 5, 6, 6, 7, 6, 7, 7, 8};


	public int bitLength()
	{
		if (nBitLength == -1) {
			if (sign == 0)
            {
                nBitLength = 0;
            }
            else if (magnitude.length == 0)
            {
                return 0;
            }
			else
            {
				// bit length for everything after the first int
				nBitLength = 32 * (magnitude.length - 1);

				// and determine bitlength of first int
				nBitLength += bitLen(magnitude[0]);

				if (sign < 0) {
					// Check if magnitude is a power of two
					boolean pow2 =
						((bitCounts[magnitude[0] & 0xff]) +
						(bitCounts[(magnitude[0] >> 8) & 0xff]) +
						(bitCounts[(magnitude[0] >> 16) & 0xff]) +
						(bitCounts[(magnitude[0] >> 24) & 0xff])) == 1;
					for (int i=1; i < magnitude.length && pow2; i++)
						pow2 = (magnitude[i] == 0);

					nBitLength -= (pow2 ? 1 : 0);
					}
				}
			}

		return nBitLength;
	}


	//
	// bitLen(val) is the number of bits in val.
	//
	static int bitLen(int w) {
        	// Binary search - decision tree (5 tests, rarely 6)
        return
         (w < 1<<15 ?
          (w < 1<<7 ?
           (w < 1<<3 ?
            (w < 1<<1 ? (w < 1<<0 ? (w<0 ? 32 : 0) : 1) : (w < 1<<2 ? 2 : 3)) :
            (w < 1<<5 ? (w < 1<<4 ? 4 : 5) : (w < 1<<6 ? 6 : 7))) :
           (w < 1<<11 ?
            (w < 1<<9 ? (w < 1<<8 ? 8 : 9) : (w < 1<<10 ? 10 : 11)) :
            (w < 1<<13 ? (w < 1<<12 ? 12 : 13) : (w < 1<<14 ? 14 : 15)))) :
          (w < 1<<23 ?
           (w < 1<<19 ?
            (w < 1<<17 ? (w < 1<<16 ? 16 : 17) : (w < 1<<18 ? 18 : 19)) :
            (w < 1<<21 ? (w < 1<<20 ? 20 : 21) : (w < 1<<22 ? 22 : 23))) :
           (w < 1<<27 ?
            (w < 1<<25 ? (w < 1<<24 ? 24 : 25) : (w < 1<<26 ? 26 : 27)) :
            (w < 1<<29 ? (w < 1<<28 ? 28 : 29) : (w < 1<<30 ? 30 : 31)))));
    }


    private final static byte bitLengths[] = {
		0, 1, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4,
		5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
		6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
		6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
		7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
		7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
		7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
		7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
		8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
		8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
		8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
		8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
		8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
		8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
		8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
		8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8};


	public int compareTo(Object o) {
		return compareTo((BigInteger)o);
	}

	public int compareTo(
		BigInteger val)
	{
		if (sign < val.sign) return -1;
		if (sign > val.sign) return 1;

		// signs the same, compare magnitude lengths
		if (magnitude.length < val.magnitude.length) return -1;
		if (magnitude.length > val.magnitude.length) return 1;

		// lengths of magnitudes the same, test the magnitude values
		for (int i = 0; i < magnitude.length; i++) {
			long v1 = (long)(magnitude[i]) & IMASK;
			long v2 = (long)(val.magnitude[i]) & IMASK;
			if (v1 < v2) return -1;
			if (v1 > v2) return 1;
			}

		return 0;
	}


	public BigInteger divide(
		BigInteger val)
		throws ArithmeticException
	{
		if (val.sign == 0) {
			throw new ArithmeticException("Divide by zero");
			}

		if (sign == 0) return BigInteger.ZERO;
		if (val.compareTo(BigInteger.ONE) == 0) return this;

		// make the players positive
		BigInteger remainder = this.abs();
		BigInteger divisor = val.abs();

		BigInteger quotient = BigInteger.ZERO;

		while (remainder.compareTo(divisor) >= 0) {
			BigInteger count = BigInteger.ONE;
			BigInteger c;
			BigInteger cc;

			int shift;
			shift = remainder.bitLength() - divisor.bitLength();
			if (shift > 0) {
				c = divisor.shiftLeft(shift-1);
				count = count.shiftLeft(shift-1);
				}
			else c = divisor;
			cc = c.shiftLeft(1);
			while (cc.compareTo(remainder) < 0) {
				count = count.shiftLeft(1);
				c = cc;
				cc = cc.shiftLeft(1);
				}

			remainder = remainder.subtract(c);
			quotient = quotient.add(count);
			}

		quotient.sign = this.sign * val.sign;

		return quotient;
	}

	public BigInteger[] divideAndRemainder(
		BigInteger val)
		throws ArithmeticException
	{
		if (val.sign == 0) {
			throw new ArithmeticException("Divide by zero");
			}

		BigInteger biggies[] = new BigInteger[2];

		if (sign == 0) {
			biggies[0] = biggies[1] = BigInteger.ZERO;
			return biggies;
			}

		if (val.compareTo(BigInteger.ONE) == 0) {
			biggies[0] = this;
			biggies[1] = BigInteger.ZERO;
			return biggies;
			}

		// make the players positive
		BigInteger remainder = this.abs();
		BigInteger divisor = val.abs();

		BigInteger quotient = BigInteger.ZERO;

		while (remainder.compareTo(divisor) >= 0) {
			BigInteger count = BigInteger.ONE;
			BigInteger c;
			BigInteger cc;

			int shift;
			shift = remainder.bitLength() - divisor.bitLength();
			if (shift > 0) {
				c = divisor.shiftLeft(shift-1);
				count = count.shiftLeft(shift-1);
				}
			else c = divisor;
			cc = c.shiftLeft(1);
			while (cc.compareTo(remainder) < 0) {
				count = count.shiftLeft(1);
				c = cc;
				cc = cc.shiftLeft(1);
				}

			remainder = remainder.subtract(c);
			quotient = quotient.add(count);
			}

		quotient.sign = this.sign * val.sign;
		if (remainder.sign != 0) remainder.sign = this.sign;
		biggies[0] = quotient;
		biggies[1] = remainder;

		return biggies;
	}


	public boolean equals(
		Object val)
	{
		if (val == this) return true;

		if (!(val instanceof BigInteger)) return false;
		BigInteger biggie = (BigInteger)val;

		if (biggie.sign != sign || biggie.magnitude.length != magnitude.length)
			return false;

		for (int i = 0; i < magnitude.length; i++) {
			if (biggie.magnitude[i] != magnitude[i]) return false;
			}

		return true;
	}

	public BigInteger gcd(
		BigInteger val)
	{
		if (val.sign == 0) return this.abs();
		else if (sign == 0) return val.abs();

		BigInteger r;
		BigInteger u = this;
		BigInteger v = val;

		while (v.sign != 0) {
			r = u.mod(v);
			u = v;
			v = r;
		}

		return u;
	}


	public int hashCode()
	{
		return 0;
	}


	public int intValue()
	{
		return getMagInt(0);
	}


    /**
     * return whether or not a BigInteger is probably prime with a 
     * probability of 1 - (1/2)**certainty.
     * <p>
     * From Knuth Vol 2, pg 395.
     */
	public boolean isProbablePrime(
		int certainty)
	{
        if (certainty == 0)
        {
            return true;
        }
        
        BigInteger n = this.abs();

        if (n.equals(TWO))
        {
            return true;
        }

        if (n.equals(ONE) || !n.testBit(0))
        {
            return false;
        }

	if ((certainty & 0x1) == 1)
	{
		certainty = certainty / 2 + 1;
	}
	else
	{
		certainty /= 2;
	}

        //
        // let n = 1 + 2^kq
        //
        BigInteger  q = n.subtract(ONE);
        int         k = q.getLowestSetBit();

        q = q.shiftRight(k);

        Random  rnd = new Random();
        for (int i = 0; i <= certainty; i++)
        {
            BigInteger x;

            do
            {
                x = new BigInteger(n.bitLength(), rnd);
            }
            while (x.compareTo(ONE) <= 0 || x.compareTo(n) >= 0);

            int         j = 0;
            BigInteger  y = x.modPow(q, n);

            while (!((j == 0 && y.equals(ONE)) || y.equals(n.subtract(ONE))))
            {
                if (j > 0 && y.equals(ONE))
                {
                    return false;
                }
                if (++j == k)
                {
                    return false;
                }
                y = y.modPow(TWO, n);
            }
        }
        
		return true;
	}


	public long longValue()
	{
		return ((long)getMagInt(1) << 32) | ((long)getMagInt(0) & IMASK);
	}


	public BigInteger max(
		BigInteger val)
	{
		return (compareTo(val) > 0) ? this : val;
	}


	public BigInteger min(
		BigInteger val)
	{
		return (compareTo(val) < 0) ? this : val;
	}


	public BigInteger mod(
		BigInteger m)
		throws ArithmeticException
	{
		if (m.sign <= 0) throw new ArithmeticException("BigInteger: modulus is not positive");
		BigInteger biggie = this.remainder(m);

		return (biggie.sign >= 0 ? biggie : biggie.add(m));
	}


	public BigInteger modInverse(
		BigInteger m)
		throws ArithmeticException
	{
//debug = true;
		if (m.sign != 1)
			throw new ArithmeticException("Modulus must be positive");

        BigInteger x = new BigInteger();
        BigInteger y = new BigInteger();

        BigInteger gcd = BigInteger.extEuclid(this, m, x, y);

        if (!gcd.equals(BigInteger.ONE))
        {
                throw new ArithmeticException("Numbers not relatively prime.");
        }

        if (x.compareTo(BigInteger.ZERO) < 0)
        {
            x = x.add(m);
        }

        return x;
	}

	/**
	 * Calculate the numbers u1, u2, and u3 such that:
	 *
	 * u1 * a + u2 * b = u3
	 *
	 * where u3 is the greatest common divider of a and b.
	 * a and b using the extended Euclid algorithm (refer p. 323
	 * of The Art of Computer Programming vol 2, 2nd ed).
	 * This also seems to have the side effect of calculating
	 * some form of multiplicative inverse.
	 *
	 * @param a    First number to calculate gcd for
	 * @param b    Second number to calculate gcd for
	 * @param u1Out      the return object for the u1 value
	 * @param u2Out      the return object for the u2 value
	 * @return     The greatest common divisor of a and b
	 */
	private static BigInteger extEuclid(
		BigInteger a,
		BigInteger b,
		BigInteger u1Out,
		BigInteger u2Out)
	{
		BigInteger  res;

		BigInteger u1 = BigInteger.ONE;
		BigInteger u3 = a;
		BigInteger v1 = BigInteger.ZERO;
		BigInteger v3 = b;

		while (v3.compareTo(BigInteger.ZERO) > 0)
		{
			BigInteger q, tn, tv;

			q = u3.divide(v3);

			tn = u1.subtract(v1.multiply(q));
			u1 = v1;
			v1 = tn;

			tn = u3.subtract(v3.multiply(q));
			u3 = v3;
			v3 = tn;
		}

		u1Out.sign = u1.sign;
		u1Out.magnitude = u1.magnitude;

		res = u3.subtract(u1.multiply(a)).divide(b);
		u2Out.sign = res.sign;
		u2Out.magnitude = res.magnitude;

		return u3;
	}

	public BigInteger modPow(
		BigInteger exponent,
		BigInteger m)
		throws ArithmeticException
	{
		BigInteger y = BigInteger.ONE;
		BigInteger z = this;

		// from LSW to MSW
		for (int i = exponent.magnitude.length - 1; i >= 0; i--)
        {
			long v = ((int)exponent.magnitude[i]) & IMASK;
			int bits = 0;

			while (v != 0) {
				if ((v & 0x1) != 0) {
					y = y.multiply(z).remainder(m);
                }

				z = z.multiply(z).remainder(m);
				v >>>= 1;
				bits++;
            }

			// not in last word, finish off the bits in this word
			if (i != 0)
            {
				while (bits < 32)
                {
					z = z.multiply(z).remainder(m);
					bits++;
                }
            }
        }

		return y;
	}


	public BigInteger multiply(
		BigInteger val)
	{
		if (sign == 0 || val.sign == 0) return BigInteger.ZERO;

		int[] mag = new int[magnitude.length + val.magnitude.length];
		int i, j;
		long t, k;

		j = 0;
		while (j < val.magnitude.length) {
			i = 0;
			k = 0;
			while (i < magnitude.length) {
				t = (((long)getMagInt(magnitude.length-i-1)) & IMASK)
				  * (((long)val.getMagInt(val.magnitude.length-j-1)) & IMASK)
				  + (((long)mag[mag.length-(i+j)-1]) & IMASK)
				  + k;
				mag[mag.length-(i+j)-1] = (int)(t & IMASK);
				k = (t >> 32) & IMASK;
				i++;
				}

			mag[mag.length-(j+magnitude.length)-1] = (int)k;
			j++;
			}

		return new BigInteger(sign * val.sign, mag);
	}


	public BigInteger negate()
	{
		return new BigInteger(-sign, magnitude);
	}


	public BigInteger pow(
		int exp)
		throws ArithmeticException
	{
		if (exp < 0) throw new ArithmeticException("Negative exponent");
		if (sign == 0) return (exp == 0 ? BigInteger.ONE : this);

		BigInteger y, z;
		y = BigInteger.ONE;
		z = this;

		while (exp != 0) {
			if ((exp & 0x1) == 1) y = y.multiply(z);
			exp >>= 1;
			if (exp != 0) z = z.multiply(z);
			}

		return y;
	}


	public BigInteger remainder(
		BigInteger val)
		throws ArithmeticException
	{
		if (val.sign == 0) {
			throw new ArithmeticException("BigInteger: Divide by zero");
			}

		if (sign == 0) return BigInteger.ZERO;

		// make the players positive
		BigInteger remainder = this.abs();
		BigInteger divisor = val.abs();

		while (remainder.compareTo(divisor) >= 0) {
			BigInteger c;
			BigInteger cc;

			int shift;
			shift = remainder.bitLength() - divisor.bitLength();
			if (shift > 0) c = divisor.shiftLeft(shift-1);
			else c = divisor;
			cc = c.shiftLeft(1);
			while (cc.compareTo(remainder) < 0) {
				c = cc;
				cc = cc.shiftLeft(1);
				}

			remainder = remainder.subtract(c);
			}

		if (remainder.sign != 0) remainder.sign = sign;

		return remainder;
	}


	public BigInteger shiftLeft(
		int n)
	{
		if (sign == 0 || magnitude.length == 0)
        {
            return ZERO;
        }
		if (n == 0) return this;
		if (n < 0) return shiftRight(-n);

		int nInts = n >>> 5;
		int nBits = n & 0x1f;
		int magLen = magnitude.length;
		int newMag[] = null;

		if (nBits == 0) {
			newMag = new int[magLen + nInts];
			for (int i = 0; i < magLen; i++) {
				newMag[i] = magnitude[i];
				}
			}
		else {
			int i = 0;
			int nBits2 = 32 - nBits;
			int highBits = magnitude[0] >>> nBits2;
			if (highBits != 0) {
				newMag = new int[magLen + nInts + 1];
				newMag[i++] = highBits;
				}
			else {
				newMag = new int[magLen + nInts];
				}
			int j=0;
			while (j < magLen-1) {
				newMag[i++] = magnitude[j++] << nBits | magnitude[j] >>> nBits2;
				}

			newMag[i] = magnitude[j] << nBits;
			}

	return new BigInteger(sign, newMag);
	}


	public BigInteger shiftRight(
		int n)
	{
		if (n == 0)
        {
            return this;
        }

		if (n < 0)
        {
            return shiftLeft(-n);
        }

		if (n >= bitLength())
        {
			return (this.sign < 0 ? valueOf(-1) : BigInteger.ZERO);
		}

        BigInteger  v = this.divide(TWO);

        for (int i = 1; i != n; i++)
        {
            v = v.divide(TWO);
        }

		return v;
	}


	public int signum()
	{
		return sign;
	}


	public BigInteger subtract(
		BigInteger val)
	{
		if (val.sign == 0 || val.magnitude.length == 0)
        {
            return this;
        }
		if (sign == 0 || magnitude.length == 0)
        {
		    return val.negate();
        }
		if (val.sign < 0)
        {
			if (this.sign > 0) return this.add(val.negate());
        }
		else
        {
			if (this.sign < 0) return this.add(val.negate());
        }

		BigInteger bigun, littlun;
		int compare = compareTo(val);
		if (compare == 0) return ZERO;
		if (compare < 0) {
			bigun = val;
			littlun = this;
			}
		else {
			bigun = this;
			littlun = val;
			}

		int mag[] = new int[max(bigun.magnitude.length, littlun.magnitude.length)];
		int i = mag.length - 1;
		int iT = bigun.magnitude.length - 1;
		int iV = littlun.magnitude.length - 1;
		long m;
		int borrow = 0;

		do {
			m =   (((long)bigun.getMagInt(iT)) & IMASK)
				- (((long)littlun.getMagInt(iV)) & IMASK)
				+ borrow;
			mag[i] = (int)m;
			if (m < 0) borrow = -1;
			else borrow = 0;
			i--;
			iV--;
			iT--;
		} while (i >= 0);

		return new BigInteger(this.sign * compare, mag);
	}


	public byte[] toByteArray()
	{
		byte[] bytes = new byte[bitLength()/8 + 1];

		int bytesCopied = 4;
		int mag = 0;
		int ofs = magnitude.length-1;
		for (int i = bytes.length-1; i >= 0; i--) {
			if (bytesCopied == 4) {
				mag = getMagInt(ofs--);
				bytesCopied = 1;
				}
			else {
				mag >>>= 8;
				bytesCopied++;
				}

			bytes[i] = (byte)mag;
			}

		return bytes;
	}


	public String toString()
	{
		return toString(10);
	}


	public String toString(int rdx)
	{
		// only radix 16 at the moment
		if (magnitude == null) {
			return "null";
			}
		else if (sign == 0) {
			return "0";
			}

		String s = sign == -1 ? "-" : "";
		String h;
		
		for (int i = 0; i < magnitude.length; i++) {
			h = "0000000" + Integer.toHexString(magnitude[i]);
			h = h.substring(h.length()-8);
			s = s + h;
			}

		while (s.length() > 1 && s.charAt(0) == '0') s = s.substring(1);

		if (s.length() == 0) s = "0";

		return s;
	}


	private static final int MAX_BIGUNS = 16;
	private static BigInteger bigunsPos[] = new BigInteger[MAX_BIGUNS+1];
	private static BigInteger bigunsNeg[] = new BigInteger[MAX_BIGUNS+1];

	static {
		for (int i = 1; i <= MAX_BIGUNS; i++) {
			byte b[] = new byte[1];
			b[0] = (byte) i;
			bigunsPos[i] = new BigInteger(1, b);
			bigunsNeg[i] = new BigInteger(-1, b);
			}
	}


	public static final BigInteger ZERO = new BigInteger(0, new byte[0]);
	public static final BigInteger ONE = valueOf(1);
	private static final BigInteger TWO = valueOf(2);


	public static BigInteger valueOf(
		long val)
	{
		if (val == 0) return BigInteger.ZERO;
		if (val > -MAX_BIGUNS && val < MAX_BIGUNS) {
			if (val > 0) return bigunsPos[(int)val];
			return bigunsNeg[(int)-val];
			}

		// store val into a byte array
		byte[] b = new byte[8];
		for (int i = 0; i < 8; i++) {
			b[7 - i] = (byte)val;
			val >>= 8;
			}

		return new BigInteger(b);
	}


	private int max(int a, int b) {
		if (a < b) return b;
		return a;
	}

//	public BigInteger and(
//		BigInteger val)
//	{
//		int[] mag = new int[max(magnitude.length, val.magnitude.length)];
//		for (int i = 0; i < mag.length; i++) {
//			mag[i] = getMagInt(mag.length - i - 1) &
//				val.getMagInt(mag.length - i - 1);
//			}
//
//		// FIXME:
//		return null; //return new BigInteger(mag);
//	}

//	public BigInteger andNot(
//		BigInteger val)
//	{
//		int[] mag = new int[max(magnitude.length, val.magnitude.length)];
//		for (int i = 0; i < mag.length; i++) {
//			mag[i] = getMagInt(mag.length - i - 1) &
//				~val.getMagInt(mag.length - i - 1);
//			}
//
//		// FIXME:
//		return null; //return new BigInteger(mag);
//	}

//	/* Make me a BigInteger with the same contents as this
//	 * but include space for 'atleastNBits' bits.
//	 * Helper routine for bit flipp'n stuff.
//	 */
//	private BigInteger duplicate(int atleastNBits) {
//		if (atleastNBits <= 0) return new BigInteger(this.sign, this.magnitude);
//
//		int nInts = atleastNBits / 32;
//		if ((atleastNBits % 32) != 0) nInts++;
//		int[] mag = new int[max(nInts, this.magnitude.length)];
//
//		// copy the ints
//		int m = this.magnitude.length - 1;
//		for (int i = mag.length-1; i >= 0; i--, m--) {
//			mag[i] = this.getMagInt(m);
//			}
//
//		return new BigInteger(this.sign, mag);
//	}

//	public BigInteger clearBit(
//		int n)
//		throws ArithmeticException
//	{
//		if (n < 0) throw new ArithmeticException("Bit position must not be negative");
//
//		BigInteger biggie = this.duplicate(n);
//
//		biggie.magnitude[biggie.magnitude.length - n / 32 - 1] &= ~(1 << n % 32);
//
//		return biggie;
//	}

//	public double doubleValue()
//	{
//		return 0.0;
//	}

//	public BigInteger flipBit(
//		int pos)
//		throws ArithmeticException
//	{
//		if (pos < 0) throw new ArithmeticException("Bit position must not be negative");
//
// 		BigInteger biggie = this.duplicate(pos);
//
//		biggie.magnitude[biggie.magnitude.length - pos / 32 - 1] ^= (1 << pos % 32);
//
//		return biggie;
//	}

//	public float floatValue()
//	{
//		return (float)0.0;
//	}

	public int getLowestSetBit()
	{
        if (this.equals(ZERO))
        {
            return -1;
        }

        int w = magnitude.length - 1;

        while (w >= 0)
        {
            if (magnitude[w] != 0)
            {
                break;
            }

            w--;
        }

        int b = 31;

        while (b > 0)
        {
            if ((magnitude[w] << b) == 0x80000000)
            {
                break;
            }

            b--;
        }

		return (((magnitude.length - 1) - w) * 32 + (31 - b));
	}

//	public BigInteger not()
//	{
//		int[] mag = new int[magnitude.length];
//
//		for (int i = 0; i < mag.length; i++) {
//			mag[i] = ~getMagInt(i);
//			}
//
//		// FIXME:
//		return null; //new BigInteger(mag);
//	}

//	public BigInteger or(
//		BigInteger val)
//	{
//		int[] mag = new int[max(magnitude.length, val.magnitude.length)];
//		for (int i = 0; i < mag.length; i++) {
//			mag[i] = getMagInt(mag.length - i - 1) |
//				val.getMagInt(mag.length - i - 1);
//			}
//
//		// FIXME:
//		return null; //new BigInteger(mag);
//	}

//	public BigInteger setBit(
//		int n)
//		throws ArithmeticException
//	{
//		if (n < 0) throw new ArithmeticException("Bit position must not be negative");
//
// 		BigInteger biggie = this.duplicate(n);
//
//		biggie.magnitude[biggie.magnitude.length - n / 32 - 1] |= (1 << n % 32);
//
//		return biggie;
//	}

	public boolean testBit(
		int n)
		throws ArithmeticException
	{
		if (n < 0) 
        {
            throw new ArithmeticException("Bit position must not be negative");
        }

        if ((n / 32) >= magnitude.length)
        {
            return sign < 0;
        }

		return ((magnitude[(magnitude.length - 1) - n / 32] >> (n % 32)) & 1) > 0;
	}

//	public BigInteger xor(
//		BigInteger val)
//	{
//		int[] mag = new int[max(magnitude.length, val.magnitude.length)];
//		for (int i = 0; i < mag.length; i++) {
//			mag[i] = getMagInt(mag.length - i - 1) ^
//				val.getMagInt(mag.length - i - 1);
//			}
//
//		// FIXME:
//		return null; //return new BigInteger(mag);
//	}

//	private byte getByte(
//		int index)
//	{
//		if (index/4 >= this.magnitude.length || index < 0) {
//			return (byte)((this.sign >= 0) ? 0 : -1);
//			}
//
//		int offset = index % 4;
//		int mag = magnitude[index/4];
//
//		if (offset == 0) return (byte)((mag >> 24) & 0xff);
//		else if (offset == 1) return (byte)((mag >> 16) & 0xff);
//		else if (offset == 2) return (byte)((mag >> 8) & 0xff);
//		return (byte)(mag & 0xff);
//	}

}
