package org.bouncycastle.asn1.x509;

import java.io.*;
import java.util.*;

import org.bouncycastle.asn1.*;

public class X509Name
    implements DEREncodable
{
    /**
     * country code - StringType(SIZE(2))
     */
    public static final DERObjectIdentifier C = new DERObjectIdentifier("2.5.4.6");

    /**
     * organization - StringType(SIZE(1..64))
     */
    public static final DERObjectIdentifier O = new DERObjectIdentifier("2.5.4.10");

    /**
     * organizational unit name - StringType(SIZE(1..64))
     */
    public static final DERObjectIdentifier OU = new DERObjectIdentifier("2.5.4.11");

    /**
     * common name - StringType(SIZE(1..64))
     */
    public static final DERObjectIdentifier CN = new DERObjectIdentifier("2.5.4.3");

    /**
     * device serial number name - StringType(SIZE(1..64))
     */
    public static final DERObjectIdentifier SN = new DERObjectIdentifier("2.5.4.5");

    /**
     * locality name - StringType(SIZE(1..64))
     */
    public static final DERObjectIdentifier L = new DERObjectIdentifier("2.5.4.7");

    /**
     * state, or province name - StringType(SIZE(1..64))
     */
    public static final DERObjectIdentifier ST = new DERObjectIdentifier("2.5.4.8");

    /**
     * email address (RSA PKCS#9 extension) - IA5String
     * <p>
     * note: if you're trying to be ultra orthodox, don't use this! It shouldn't be in here.
     */
    public static final DERObjectIdentifier EmailAddress = new DERObjectIdentifier("1.2.840.113549.1.9.1");

    /**
     * look up table translating OID values into their common symbols.
     */
    public static Hashtable OIDLookUp = new Hashtable();

    static
    {
        OIDLookUp.put(C, "C");
        OIDLookUp.put(O, "O");
        OIDLookUp.put(OU, "OU");
        OIDLookUp.put(CN, "CN");
        OIDLookUp.put(L, "L");
        OIDLookUp.put(ST, "ST");
        OIDLookUp.put(SN, "SN");
        OIDLookUp.put(EmailAddress, "EmailAddress");
    }

    private Hashtable               attributes = new Hashtable();
    private DERConstructedSequence  seq;

    /**
     * Constructor from DERConstructedSequence.
     *
     * the principal will be a list of constructed sets, each containing an (OID, String) pair.
     */
    public X509Name(
        DERConstructedSequence  seq)
    {
        this.seq = seq; 

        Enumeration e = seq.getObjects();

        while (e.hasMoreElements())
        {
            DERSet  set = (DERSet)e.nextElement();
            DERConstructedSequence  s = (DERConstructedSequence)set.getSequence();

            attributes.put(s.getObjectAt(0), s.getObjectAt(1));
        }
    }

    /**
     * constructor from a table of attributes.
     * <p>
     * it's is assumed the table contains OID/String pairs, and the contents
     * of the table are copied into an internal table as part of the 
     * construction process.
     */
    public X509Name(
        Hashtable  attributes)
    {
        this.seq = new DERConstructedSequence(); 

        Enumeration     e = attributes.keys();

        while (e.hasMoreElements())
        {
            DERObjectIdentifier     oid = (DERObjectIdentifier)e.nextElement();
            DERConstructedSequence  s = new DERConstructedSequence();

            if (OIDLookUp.get(oid) == null)
            {
                throw new IllegalArgumentException("Unknown object id - " + oid.getId() + " - passed to distinguished name");
            }

            s.addObject(oid);
            if (oid.equals(EmailAddress))
            {
                s.addObject(new DERIA5String((String)attributes.get(oid)));
            }
            else
            {
                s.addObject(new DERPrintableString((String)attributes.get(oid)));
            }

            seq.addObject(new DERSet(s));

            this.attributes.put(oid, attributes.get(oid)); // copy the hash table
        }
    }

    public DERObject getDERObject()
    {
        return seq;
    }

    public int hashCode()
    {
        Enumeration     e = attributes.keys();
        int             hashCode = 0;

        while (e.hasMoreElements())
        {
            Object  o = e.nextElement();

            hashCode ^= o.hashCode();
            hashCode ^= attributes.get(o).hashCode();
        }

        return hashCode;
    }

    public boolean equals(
        Object o)
    {
        if (o == null || !(o instanceof X509Name))
        {
            return false;
        }

        X509Name        other = (X509Name)o;

        Enumeration     e1 = attributes.keys();
        Enumeration     e2 = other.attributes.keys();

        while (e1.hasMoreElements() && e2.hasMoreElements())
        {
            Object  o1 = e1.nextElement();
            Object  o2 = e2.nextElement();
            
            if (!o1.equals(o2))
            {
                return false;
            }
        }

        if (e1.hasMoreElements() || e2.hasMoreElements())
        {
            return false;
        }

        return true;
    }

    public String toString()
    {
        Enumeration     e = attributes.keys();
        boolean         first = true;
        StringBuffer    buf = new StringBuffer();

        while (e.hasMoreElements())
        {
            Object  oid = e.nextElement();
            String  sym = (String)OIDLookUp.get(oid);
            
            if (first)
            {
                first = false;
            }
            else
            {
                buf.append(", ");
            }

            if (sym != null)
            {
                buf.append(sym);
                buf.append("=");
                buf.append(((DERString)attributes.get(oid)).getString());
            }
            else
            {
                buf.append(((DERObjectIdentifier)oid).getId());
                buf.append("=");
                buf.append(((DERString)attributes.get(oid)).getString());
            }
        }

        return buf.toString();
    }
}
